/* acrostic-generator.c
 *
 * Copyright 2023-2024 Tanmay Patil <tanmaynpatil105@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#include "acrostic-generator.h"
#include "word-list.h"


G_DEFINE_QUARK (acrostic-generator-error-quark, acrostic_generator_error)

#define MIN_WORD_SIZE 3
#define MAX_WORD_SIZE 20

enum
{
  PROP_0,
  PROP_RANDOM_SEED,
  PROP_MIN_WORD_SIZE,
  PROP_MAX_WORD_SIZE,
  PROP_QUOTE_STR,
  PROP_SOURCE_STR,
  PROP_CHARSET,
  PROP_WORD_LIST_RESOURCE,
  PROP_SKIP_TABLE,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] = {NULL, };


struct _AcrosticGenerator
{
  GObject parent_object;
  guint32 random_seed;
  IpuzCharset *charset;

  gchar *quote_str;
  gchar *source_str;

  /* Clues */
  GArray *clues;

  /* Word List */
  WordListResource *word_list_resource;
  GHashTable *skip_table;

  /* Word size */
  guint max_word_size;
  guint min_word_size;

  /* Grid */
  guint n_rows;
  guint n_columns;

  GCancellable *cancellable;

  gint *count;
  GMutex *results_mutex;
  GArray *results;
};

static void acrostic_generator_init         (AcrosticGenerator      *self);
static void acrostic_generator_class_init   (AcrosticGeneratorClass *klass);
static void acrostic_generator_set_property (GObject                *object,
                                             guint                   prop_id,
                                             const GValue           *value,
                                             GParamSpec             *pspec);
static void acrostic_generator_get_property (GObject                *object,
                                             guint                   prop_id,
                                             GValue                 *value,
                                             GParamSpec             *pspec);
static void acrostic_generator_finalize     (GObject                *object);
static void acrostic_generator_set_run_info (PuzzleTask             *self,
                                             gint                   *count,
                                             GMutex                 *results_mutex,
                                             GArray                 *results);
static void acrostic_generator_run          (GTask                  *task,
                                             GObject                *source_object,
                                             gpointer                task_data,
                                             GCancellable           *cancellable);


G_DEFINE_TYPE (AcrosticGenerator, acrostic_generator, PUZZLE_TYPE_TASK);

static void
acrostic_generator_init (AcrosticGenerator *self)
{
  self->max_word_size = MAX_WORD_SIZE;
  self->min_word_size = MIN_WORD_SIZE;
}

static void
acrostic_generator_class_init (AcrosticGeneratorClass *klass)
{
  GObjectClass *object_class;
  PuzzleTaskClass *task_class;

  object_class = G_OBJECT_CLASS (klass);
  task_class = PUZZLE_TASK_CLASS (klass);

  object_class->set_property = acrostic_generator_set_property;
  object_class->get_property = acrostic_generator_get_property;
  object_class->finalize = acrostic_generator_finalize;
  task_class->run = acrostic_generator_run;
  task_class->set_run_info = acrostic_generator_set_run_info;

  obj_props[PROP_RANDOM_SEED] = g_param_spec_uint ("random-seed",
                                                   "Random seed",
                                                   "Seed value for rand()",
                                                   0,
                                                   G_MAXUINT32,
                                                   0,
                                                   G_PARAM_READWRITE);

  obj_props[PROP_MIN_WORD_SIZE] = g_param_spec_uint ("min-word-size",
                                                     "Minimum word size",
                                                     "Minimum size of a clue word",
                                                     MIN_WORD_SIZE,
                                                     MAX_WORD_SIZE,
                                                     MIN_WORD_SIZE,
                                                     G_PARAM_READWRITE);

  obj_props[PROP_MAX_WORD_SIZE] = g_param_spec_uint ("max-word-size",
                                                     "Maximum word size",
                                                     "Maximum size of a clue word",
                                                     MIN_WORD_SIZE,
                                                     MAX_WORD_SIZE,
                                                     MAX_WORD_SIZE,
                                                     G_PARAM_READWRITE);

  obj_props[PROP_QUOTE_STR] = g_param_spec_string ("quote-string",
                                                   "Quote string",
                                                   "Quote string",
                                                   NULL,
                                                   G_PARAM_READWRITE);

  obj_props[PROP_SOURCE_STR] = g_param_spec_string ("source-string",
                                                    "Source string",
                                                    "Source string",
                                                    NULL,
                                                    G_PARAM_READWRITE);

  obj_props[PROP_CHARSET] = g_param_spec_boxed ("charset",
                                                NULL, NULL,
                                                IPUZ_TYPE_CHARSET,
                                                G_PARAM_READWRITE);

  obj_props[PROP_WORD_LIST_RESOURCE] =
    g_param_spec_object ("word-list-resource",
                         NULL, NULL,
                         WORD_TYPE_LIST_RESOURCE,
                         G_PARAM_READWRITE);

  obj_props[PROP_SKIP_TABLE] = g_param_spec_boxed ("skip-table",
                                                   NULL, NULL,
                                                   G_TYPE_HASH_TABLE,
                                                   G_PARAM_READWRITE);


  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
acrostic_generator_set_property (GObject      *object,
                                 guint         prop_id,
                                 const GValue *value,
                                 GParamSpec   *pspec)
{
  AcrosticGenerator *self;

  self = ACROSTIC_GENERATOR (object);

  switch (prop_id)
    {
    case PROP_RANDOM_SEED:
      self->random_seed = g_value_get_uint (value);
      break;
    case PROP_MIN_WORD_SIZE:
      self->min_word_size = g_value_get_uint (value);
      break;
    case PROP_MAX_WORD_SIZE:
      self->max_word_size = g_value_get_uint (value);
      break;
    case PROP_QUOTE_STR:
      g_clear_pointer (&self->quote_str, g_free);
      self->quote_str = g_value_dup_string (value);
      break;
    case PROP_SOURCE_STR:
      g_clear_pointer (&self->source_str, g_free);
      self->source_str = g_value_dup_string (value);
      break;
    case PROP_CHARSET:
      if (self->charset != NULL)
        ipuz_charset_unref (self->charset);
      self->charset = g_value_dup_boxed (value);
      break;
    case PROP_WORD_LIST_RESOURCE:
      g_clear_object (&self->word_list_resource);
      self->word_list_resource = g_value_dup_object (value);
      break;
    case PROP_SKIP_TABLE:
      if (self->skip_table)
        g_hash_table_destroy (self->skip_table);
      self->skip_table = g_value_get_boxed (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
acrostic_generator_get_property (GObject    *object,
                                 guint       prop_id,
                                 GValue     *value,
                                 GParamSpec *pspec)
{
  AcrosticGenerator *self;

  g_return_if_fail (object != NULL);

  self = ACROSTIC_GENERATOR (object);

  switch (prop_id)
    {
    case PROP_RANDOM_SEED:
      g_value_set_uint (value, self->random_seed);
      break;
    case PROP_MIN_WORD_SIZE:
      g_value_set_uint (value, self->min_word_size);
      break;
    case PROP_MAX_WORD_SIZE:
      g_value_set_uint (value, self->max_word_size);
      break;
    case PROP_QUOTE_STR:
      g_value_set_string (value, self->quote_str);
      break;
    case PROP_SOURCE_STR:
      g_value_set_string (value, self->source_str);
      break;
    case PROP_CHARSET:
      g_value_set_boxed (value, self->charset);
      break;
    case PROP_WORD_LIST_RESOURCE:
      g_value_set_object (value, self->word_list_resource);
      break;
    case PROP_SKIP_TABLE:
      g_value_set_boxed (value, self->skip_table);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
acrostic_generator_finalize (GObject *object)
{
  AcrosticGenerator *self;

  self = ACROSTIC_GENERATOR (object);

  g_clear_pointer (&self->quote_str, g_free);
  g_clear_pointer (&self->source_str, g_free);
  g_clear_pointer (&self->clues, g_array_unref);
  g_clear_object (&self->cancellable);
  g_clear_object (&self->word_list_resource);
  g_clear_pointer (&self->skip_table, g_hash_table_unref);
  ipuz_charset_unref (self->charset);

  G_OBJECT_CLASS (acrostic_generator_parent_class)->finalize (object);
}

/* Public methods */

AcrosticGenerator*
acrostic_generator_new (void)
{
  GObject *object;
  IpuzCharsetBuilder *builder;
  IpuzCharset *charset;
  GHashTable *skip_table;
  GRand *g_rand;
  guint32 random_seed;

  g_rand = g_rand_new ();
  random_seed = g_rand_int (g_rand);

  /* FIXME: use the default charset */
  builder = ipuz_charset_builder_new_for_language ("C");
  charset = ipuz_charset_builder_build (builder);
  builder = NULL;
  skip_table = g_hash_table_new_full (g_str_hash, g_str_equal,
                                      g_free, NULL);

  object = g_object_new (ACROSTIC_TYPE_GENERATOR,
                         "random-seed", random_seed,
                         "charset", charset,
                         "skip-table", skip_table,
                         NULL);

  g_rand_free (g_rand);

  return ACROSTIC_GENERATOR (object);
}

void
acrostic_generator_set_min_word_size (AcrosticGenerator *self,
                                      guint              new_size)
{
  g_return_if_fail (ACROSTIC_IS_GENERATOR (self));

  self->min_word_size = new_size;
}

void
acrostic_generator_set_max_word_size (AcrosticGenerator *self,
                                      guint              new_size)
{
  g_return_if_fail (ACROSTIC_IS_GENERATOR (self));

  self->max_word_size = new_size;
}

static gboolean
check_if_possible (AcrosticGenerator *self,
                   const gchar       *quote_str,
                   const gchar       *source_str,
                   GError           **error)
{
  g_autoptr (IpuzCharset) quote_charset = NULL;
  g_autoptr (IpuzCharset) source_charset = NULL;

  quote_charset = ipuz_charset_deserialize (quote_str);
  source_charset = ipuz_charset_deserialize (source_str);

  if ((ipuz_charset_get_total_count (source_charset) * self->min_word_size) > ipuz_charset_get_total_count (quote_charset))
    {
       /*
        *  QUOTE : LIFE IS TOO SHORT
        *  SOURCE : TOLSTOI
        *
        *  min_word_size = 3 (minimun word size for each clue)
        *
        *  T _ _
        *  O _ _
        *  L _ _               quote_len = 14
        *  S _ _               source_len = 7
        *  T _ _               total_min_clue_text_len = source_len * min_word_size = 7 * 3 = 21
        *  O _ _               Since, clue_text_len > quote_len => Invalid
        *  I _ _
        *
        */
      *error = g_error_new (ACROSTIC_GENERATOR_ERROR, ACROSTIC_GENERATOR_ERROR_SOURCE_TOO_SHORT,
                            "Source text is too short to accomodate the specificed minimum word size for each clue");
      return FALSE;
    }

  for (guint i = 0; i < ipuz_charset_get_n_chars (source_charset); i++)
    {
      IpuzCharsetValue value;
      ipuz_charset_get_value (source_charset, i, &value);

      if (value.count > ipuz_charset_get_char_count (quote_charset, value.c))
        {
          *error = g_error_new (ACROSTIC_GENERATOR_ERROR, ACROSTIC_GENERATOR_ERROR_SOURCE_NOT_IN_QUOTE,
                                "Source characters are missing in the provided quote");
          return FALSE;
        }
    }

  return TRUE;
}

/*
 * normalise_input
 *
 * Remove spaces, punctuations from the string
 * and convert it to uppercase
 */
static gchar*
normalise_input (const gchar *str,
                 IpuzCharset *charset)
{
  GString *normalised;
  const char *p;

  normalised = g_string_new (NULL);

  for (p = str; p[0]; p = g_utf8_next_char (p))
    {
      gunichar c = g_unichar_toupper (g_utf8_get_char (p));

      if (ipuz_charset_get_char_count (charset, c) > 0)
        g_string_append_unichar (normalised, c);
    }

  return g_string_free_and_steal (normalised);
}

gboolean
acrostic_generator_set_text (AcrosticGenerator *self,
                             const gchar       *quote_str,
                             const gchar       *source_str,
                             GError           **error)
{
  /* normalised quote and source strings */
  gchar *n_quote_str, *n_source_str;

  g_return_val_if_fail (quote_str != NULL, FALSE);
  g_return_val_if_fail (source_str != NULL, FALSE);

  n_quote_str = normalise_input (quote_str, self->charset);
  n_source_str = normalise_input (source_str, self->charset);

  if (!check_if_possible (self, n_quote_str, n_source_str, error))
    {
      g_free (n_quote_str);
      g_free (n_source_str);

      /* callee function sets the error */
      return FALSE;
    }

  /* clear old values */
  g_clear_pointer (&self->quote_str, g_free);
  g_clear_pointer (&self->source_str, g_free);

  self->quote_str = n_quote_str;
  self->source_str = n_source_str;

  return TRUE;
}

void
acrostic_generator_set_seed (AcrosticGenerator *self,
                             guint32            seed)
{
  g_return_if_fail (ACROSTIC_IS_GENERATOR (self));

  self->random_seed = seed;
}

/* Answer entry */
typedef struct {
  gunichar leading_char;
  const gchar *letters;
  guint word_length;
  gchar *filter;
  WordList *word_list;
  GArray *rand_offset;
} ClueEntry;

typedef struct ClueScore {
  guint index;
  guint score;
} ClueScore;

static void
clue_entry_clear (ClueEntry *clue_entry)
{
  g_clear_pointer (&clue_entry->filter, g_free);
  g_clear_object (&clue_entry->word_list);
  g_clear_pointer (&clue_entry->rand_offset, g_array_unref);
}

static void
shuffle_array (GArray *arr)
{
  if (arr->len == 0)
    return;

  for (guint i = 0; i < arr->len - 1; i++)
    {
      guint j = i + rand() / (RAND_MAX / (arr->len - i) + 1);
      gushort swap = g_array_index (arr, gushort, j);
      g_array_index (arr, gushort, j) = g_array_index (arr, gushort, i);
      g_array_index (arr, gushort, i) = swap;
    }
}

static GArray*
generate_random_lookup (guint size)
{
  GArray *array;

  array = g_array_new (FALSE, TRUE, sizeof (gushort));

  g_array_set_size (array, size);

  for (guint i = 0; i < size; i++)
    g_array_index (array, gushort, i) = (gushort) i;

  shuffle_array (array);

  return array;
}

static gboolean
add_or_skip_word (AcrosticGenerator *self,
                  const gchar       *word)
{
  if (word == NULL)
    return FALSE;

  if (g_hash_table_contains (self->skip_table, word))
    return TRUE;

  g_hash_table_insert (self->skip_table, g_strdup (word), GINT_TO_POINTER (TRUE));
  return FALSE;
}

static void
clean_up_word (AcrosticGenerator *self,
               const gchar       *word)
{
  if (word == NULL)
    return;

  g_hash_table_remove (self->skip_table, word);
}

static gchar *
generate_individual_filter (gunichar leading_char,
                            guint    len)
{
  GString *filter;

  filter = g_string_sized_new (len);

  g_string_append_unichar (filter, leading_char);

  for (guint i = 1; i < len; i++)
    g_string_append_c (filter, '?');

  return g_string_free_and_steal (filter);
}

static void
fill_clue_entries (GArray           *clues,
                   ClueScore        *candidates,
                   WordListResource *resource)
{
  for (guint i = 0; i < clues->len; i++)
    {
      ClueEntry *clue_entry;

      clue_entry = &(g_array_index (clues, ClueEntry, i));

      clue_entry->filter = generate_individual_filter (clue_entry->leading_char,
                                                       clue_entry->word_length);

      clue_entry->word_list = word_list_new ();
      word_list_set_resource (clue_entry->word_list, resource);
      word_list_set_filter (clue_entry->word_list, clue_entry->filter, WORD_LIST_MATCH);

      candidates[i].index = i;
      candidates[i].score = clue_entry->word_length;

      clue_entry->rand_offset = generate_random_lookup (word_list_get_n_items (clue_entry->word_list));
    }
}

typedef enum
{
  ACROSTIC_GENERATOR_RUNNING,
  ACROSTIC_GENERATOR_SUCCESS,
  ACROSTIC_GENERATOR_FAILURE,
} AcrosticGeneratorState;

static void
dump_clues (GArray *clues, AcrosticGeneratorState state)
{
  GString *dump_str = g_string_new (NULL);

  switch (state)
    {
      case ACROSTIC_GENERATOR_RUNNING:
        break;
      case ACROSTIC_GENERATOR_SUCCESS:
        g_string_append (dump_str, "\033[32m");
        break;
      case ACROSTIC_GENERATOR_FAILURE:
        g_string_append (dump_str, "\033[31m");
        break;
    }

  for (guint i = 0; i < clues->len; i++)
    {
      ClueEntry clue_entry;

      clue_entry = g_array_index (clues, ClueEntry, i);
      if (clue_entry.letters)
        {
          g_string_append_printf (dump_str, "[ %s ] ", clue_entry.letters);
        }
      else
        {
          g_string_append_printf (dump_str, "[ ");
          g_string_append_unichar (dump_str, clue_entry.leading_char);

          if (clue_entry.word_length > 0)
            {
              for (guint j = 0; j < clue_entry.word_length - 1; j++)
                g_string_append_c (dump_str, '*');
            }

          g_string_append (dump_str, " ] ");
        }
    }

  if (state != ACROSTIC_GENERATOR_RUNNING)
    g_string_append (dump_str, "\033[0m");

  g_print ("%s", dump_str->str);
  g_print ("\n");
  g_string_free (dump_str, TRUE);
}

static gboolean
acrostic_generator_helper (AcrosticGenerator  *self,
                           GArray             *clues,
                           guint               index,
                           IpuzCharsetBuilder *remaining_letters,
                           ClueScore          *candidates)
{
  ClueEntry *clue_entry;

  if (g_cancellable_is_cancelled (self->cancellable))
    return FALSE;

  if (index == clues->len)
    return TRUE;

  clue_entry = &(g_array_index (clues, ClueEntry, candidates[index].index));

  for (guint i = 0; i < word_list_get_n_items (clue_entry->word_list); i++)
    {
      const gchar *word;

      g_atomic_int_inc (self->count);

      word = word_list_get_word (clue_entry->word_list,
                                 g_array_index (clue_entry->rand_offset, gushort, i));

      clue_entry->letters = word;

#ifdef DEVELOPMENT_BUILD
      /* Give some progress */
      if (*self->count % 1000000 == 0)
        dump_clues (clues, ACROSTIC_GENERATOR_RUNNING);
#endif

      if (ipuz_charset_builder_remove_text (remaining_letters, word + 1))
        {
          if (!add_or_skip_word (self, word) &&
               acrostic_generator_helper (self, clues, index + 1, remaining_letters, candidates))
            return TRUE;

          if (g_cancellable_is_cancelled (self->cancellable))
            break;

          clean_up_word (self, word);
          ipuz_charset_builder_add_text (remaining_letters, word + 1);
          clue_entry->letters = NULL;
        }

    }

  clue_entry->letters = NULL;

  return FALSE;
}

static GArray*
initialize_clue_entries (const char *text)
{
  const char *p;
  GArray *clues;

  clues = g_array_new (FALSE, TRUE, sizeof (ClueEntry));

  for (p = text; p[0]; p = g_utf8_next_char (p))
    {
      gunichar c;
      ClueEntry clue_entry;

      c = g_utf8_get_char (p);

      clue_entry.leading_char = c;
      clue_entry.letters = NULL;
      clue_entry.word_length = 0;

      g_array_append_val (clues, clue_entry);
    }

  return clues;
}

static gboolean
generate_random_lengths (GArray    *clues,
                         guint      number,
                         guint      min_word_size,
                         guint      max_word_size)
{
  if ((clues->len * max_word_size) < number)
    return FALSE;

  guint sum = 0;

  for (guint i = 0; i < clues->len; i++)
    {
      ClueEntry *clue_entry;
      guint len;
      guint max_len = MAX (min_word_size,
                           MIN (max_word_size, number - sum));

      len = rand() % (max_len - min_word_size + 1) + min_word_size;
      sum += len;

      clue_entry = &(g_array_index (clues, ClueEntry, i));
      clue_entry->word_length = len;
    }

  return sum == number;
}

static guint
minimise_max_word_size (guint quote_len,
                        guint source_len,
                        guint min_word_size,
                        guint max_word_size)
{
  guint minimised;

  minimised = quote_len - (source_len - 1) * min_word_size;

  return MIN (max_word_size, minimised);
}

int
compare_candidates (const void *a,
                    const void *b)
{
  ClueScore candidate_a = * (ClueScore *) a;
  ClueScore candidate_b = * (ClueScore *) b;

  return (candidate_b.score - candidate_a.score);
}

static GArray *
clue_entry_get_answers (GArray *clues)
{
  GArray *answers = NULL;

  g_return_val_if_fail (clues != NULL, NULL);
  g_return_val_if_fail (clues->len > 0, NULL);

  answers = g_array_new (FALSE, FALSE, sizeof(gchar *));

  for (guint i = 0; i < clues->len; i++)
    {
      ClueEntry clue;

      clue = g_array_index (clues, ClueEntry, i);

      g_array_append_val (answers, clue.letters);
    }

  return answers;
}

static void
acrostic_generator_run_helper (AcrosticGenerator *self)
{
  guint max_word_size, quote_length, source_length;
  GArray *clues;
  IpuzCharsetBuilder *quote_builder;
  ClueScore *candidates;
  gboolean success;

  g_return_if_fail (ACROSTIC_IS_GENERATOR (self));

  /* Init RNG */
  srand (self->random_seed);

  /* Generate random lengths */
  clues = initialize_clue_entries (self->source_str);
  g_array_set_clear_func (clues, (GDestroyNotify) clue_entry_clear);

  quote_length = g_utf8_strlen (self->quote_str, -1);
  source_length = g_utf8_strlen (self->source_str, -1);

  max_word_size = minimise_max_word_size (quote_length, source_length,
                                          self->min_word_size, self->max_word_size);

  while (!generate_random_lengths (clues, quote_length,
                                   self->min_word_size, max_word_size));

  /* Init clue entries */
  candidates = g_new (ClueScore, source_length);

  fill_clue_entries (clues, candidates, self->word_list_resource);
  qsort (candidates, clues->len, sizeof(ClueScore), compare_candidates);

  /* Remove source string from charset in advance */
  quote_builder = ipuz_charset_builder_new ();
  ipuz_charset_builder_add_text (quote_builder, self->quote_str);

  if (G_UNLIKELY (ipuz_charset_builder_remove_text (quote_builder, self->source_str) == 0))
    {
      /* Ideally, this code block is unreachable since we verify
       * the condition during `_set_text` */
      g_clear_pointer (&clues, g_array_unref);
      ipuz_charset_unref (ipuz_charset_builder_build (quote_builder));
      g_free (candidates);

      return;
    }

  /* Run generator */
  success = acrostic_generator_helper (self, clues, 0, quote_builder, candidates);

  dump_clues (clues, success ? ACROSTIC_GENERATOR_SUCCESS : ACROSTIC_GENERATOR_FAILURE);

  if (!success)
    {
      g_clear_pointer (&clues, g_array_unref);
      clues = NULL;
    }

  /* Save to results */
  if (clues != NULL)
    {
      g_mutex_lock (self->results_mutex);

      if (!g_cancellable_is_cancelled (self->cancellable))
        {
          GArray *answers;

          answers = clue_entry_get_answers (clues);

          g_array_append_val (self->results, answers);
        }

      g_mutex_unlock (self->results_mutex);

      g_clear_pointer (&clues, g_array_unref);
    }

  /* Cleanup */
  g_free (candidates);
  ipuz_charset_unref (ipuz_charset_builder_build (quote_builder));
}

void
acrostic_generator_run (GTask        *task,
                        GObject      *source_object,
                        gpointer      task_data,
                        GCancellable *cancellable)
{
  AcrosticGenerator *self = ACROSTIC_GENERATOR (task_data);

  g_return_if_fail (ACROSTIC_IS_GENERATOR (self));

  self->cancellable = g_object_ref (cancellable);

  while (!g_cancellable_is_cancelled (self->cancellable))
    {
      acrostic_generator_run_helper (self);
      self->random_seed++;
    }

  g_task_return_pointer (task, NULL, NULL);
}

static void
acrostic_generator_set_run_info (PuzzleTask *self,
                                 gint       *count,
                                 GMutex     *results_mutex,
                                 GArray     *results)
{
  ACROSTIC_GENERATOR (self)->count = count;
  ACROSTIC_GENERATOR (self)->results_mutex = results_mutex;
  ACROSTIC_GENERATOR (self)->results = results;
}

void
acrostic_generator_set_word_resource (AcrosticGenerator *self,
                                      WordListResource  *resource)
{
  g_return_if_fail (ACROSTIC_IS_GENERATOR (self));

  g_object_set (self, "word-list-resource", resource, NULL);
}
